<?php

namespace Drupal\azure_oauth_sso\Controller;

use Drupal\azure_oauth_sso\BaseOAuth;
use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Routing\TrustedRedirectResponse;
use Drupal\Core\Url;
use Drupal\file\Entity\File;
use Drupal\user\Entity\User;
use GuzzleHttp\Client;
use GuzzleHttp\Exception\RequestException;
use Symfony\Component\HttpFoundation\RedirectResponse;

class OauthLogin extends ControllerBase {
  use BaseOAuth;
  private $authCode;
  private $rolesMapping;
  private $fieldsMapping;
  public $accessToken;
  public $refreshToken;
  public $displayName;
  public $department;

  public function __construct() {
    $this->rolesMapping = \Drupal::config('azure_oauth_sso_roles_mapping.settings');
    $this->fieldsMapping = \Drupal::config('azure_oauth_sso_fields_mapping.settings');
  }

  private function generateToken() {
    $client = new Client();

    try {
      $res = $client->post(
        'https://login.microsoftonline.com/' . $this->getAdTenant() . '/oauth2/v2.0/token', [
          'form_params' => [
            'grant_type'    => 'authorization_code',
            'client_id'     => $this->getClientId(),
            'redirect_uri'  => $this->getRedirectUri(),
            'code'          => $this->authCode,
            'client_secret' => $this->getClientSecret(),
          ],
        ]
      );

      $response = json_decode($res->getBody(), TRUE);

      if (isset($response['access_token']) && isset($response['refresh_token'])) {
        $this->accessToken = 'Bearer ' . $response['access_token'];
        $this->refreshToken = $response["refresh_token"];
      }
    }
    catch (RequestException $e) {
      if ($e->hasResponse()) {
        $statusCode = $e->getResponse()->getStatusCode();

        if ($statusCode == 401 || $statusCode == 400) {
          $path = Url::fromRoute('azure_oauth_sso.oauthLogin')->toString();
          $response = new RedirectResponse($path);
          $response->send();
          exit();
        }
      }
    }
  }

  public function testOauthLogin() {
    if (!isset($_GET["code"]) and !isset($_GET["error"])) {
      $url = "https://login.microsoftonline.com/" . $this->getAdTenant() . "/oauth2/v2.0/authorize?";
      $url .= "client_id=" . $this->getClientId();
      $url .= "&response_type=code";
      $url .= "&response_mode=query";
      $url .= "&scope=https://graph.microsoft.com/User.Read offline_access";
      $url .= "&state=12345";
      return new TrustedRedirectResponse($url);
    }
  }

  public function oauthLogin() {
    $user = \Drupal::currentUser();
    if (!isset($_GET["code"]) and !isset($_GET["error"]) && !$user->isAuthenticated()) {
      $url = "https://login.microsoftonline.com/" . $this->getAdTenant() . "/oauth2/v2.0/authorize?";
      $url .= "client_id=" . $this->getClientId();
      $url .= "&response_type=code";
      $url .= "&redirect_uri=" . urlencode($this->getRedirectUri());
      $url .= "&response_mode=query";
      $url .= "&scope=https://graph.microsoft.com/User.Read offline_access";
      $url .= "&state=12345";

      return new TrustedRedirectResponse($url);
    }

    if (isset($_GET["code"])) {
      $this->authCode = $_GET["code"];
      $userPicture = $this->fieldsMapping->get('user_picture');

      if (!$user->isAuthenticated()) {
        $this->generateToken();
        $this->loginUser();
        if ($userPicture == 1) {
          $this->getMyPhoto();
        }
      }
    }

    return new RedirectResponse(Url::fromRoute('<front>')->toString());
  }

  public function loginUser() {
    $apiRes = $this->getPersonalData();
    $uid = \Drupal::entityQuery('user')->accessCheck(TRUE)->execute();
    $users = User::loadMultiple($uid);
    $found = FALSE;

    foreach ($users as $user) {
      if ($user->getEmail() === $apiRes['mail']) {
        $found = TRUE;
        break;
      }
    }

    if ($found) {
      $fields = array_flip($this->fieldsMapping->get('custom_attr_table'));
      $fieldsKeys = array_keys($fields);
      $matchedValues = [];

      foreach ($fieldsKeys as $key) {
        if (array_key_exists($key, $apiRes)) {
          $matchedValues[$fields[$key]] = $apiRes[$key];
        }
      }

      $userValues = [];

      foreach ($matchedValues as $mappedFieldName => $fieldName) {
        if ($user->hasField($mappedFieldName)) {
          $fieldValue = $user->get($mappedFieldName)->value;
          $userValues[$mappedFieldName] = $fieldValue;
        }
      }

      $differences = array_diff_assoc($matchedValues, $userValues);

      if (!empty($differences)) {
        foreach ($differences as $fieldName => $fieldValue) {
          $user->set($fieldName, $fieldValue);
        }
        $user->save();
      }

      $user->set('field_access_token', $this->accessToken);
      $user->set('field_refresh_token', $this->refreshToken);
      $user->save();
      user_login_finalize($user);
    }
    else {
      $roleValue = $this->rolesMapping->get('custom_attr_table');
      $roleOptions = $this->rolesMapping->get('role_options');
      $defaultRole = $this->rolesMapping->get('default_role');

      if ($roleOptions == 'ad_group') {
        $id = $this->getRolesPerGroup();
        $role = array_search($id, $roleValue);
        if ($role === FALSE) {
          $role = $defaultRole;
        }
      }
      elseif ($roleOptions == 'department_field') {
        $role = array_search($this->department, $roleValue);
        if ($role === FALSE) {
          $role = $defaultRole;
        }
      }
      else {
        $role = $defaultRole;
      }

      $fields = array_flip($this->fieldsMapping->get('custom_attr_table'));
      $fieldsKeys = array_keys($fields);

      $matchedValues = [];

      foreach ($fieldsKeys as $key) {
        if (array_key_exists($key, $apiRes)) {
          $matchedValues[$fields[$key]] = $apiRes[$key];
        }
      }

      $user = User::create($matchedValues);

      $user->addRole($role);
      $user->set("status", 1);
      $user->set("field_access_token", $this->accessToken);
      $user->set("field_refresh_token", $this->refreshToken);

      $user->save();

      user_login_finalize($user);

    }
  }

  public function getPersonalData() {
    $fieldsMapping = array_values($this->fieldsMapping->get('custom_attr_table'));
    $fieldsMapping[] = 'displayName,department';

    $filteredValues = array_filter($fieldsMapping, function ($value) {
      return $value !== '_none' && !empty ($value);
    });

    $urlApi = 'https://graph.microsoft.com/v1.0/me';

    if (!empty($filteredValues)) {
      $selectQueryString = '$select=' . implode(',', $filteredValues);

      $urlApi .= '?' . $selectQueryString;
    }

    $params = [
      'headers' => [
        'Content-Type'  => 'application/json',
        'Authorization' => $this->accessToken,
      ],
    ];

    $apiRes = \Drupal::service('azure_oauth_sso.token_service')->apiCall($urlApi, $params, 'GET');
    $this->displayName = $apiRes['displayName'] ?? NULL;
    $this->department = $apiRes['department'] ?? NULL;
    return $apiRes;
  }

  public function getRolesPerGroup() {
    $urlApi = 'https://graph.microsoft.com/v1.0/me/transitiveMemberOf/microsoft.graph.group?$count=true';

    $params = [
      'headers' => [
        'Content-Type'  => 'application/json',
        'Authorization' => $this->accessToken,
      ],
    ];

    $apiRes = \Drupal::service('azure_oauth_sso.token_service')->apiCall($urlApi, $params, 'GET');
    if (!empty($apiRes["value"])) {
      foreach ($apiRes["value"] as $data) {
        $id = $data["id"];
      }
    }

    return $id;
  }

  public function getMyPhoto() {
    $token = \Drupal::service('azure_oauth_sso.token_service')->getToken();
    $urlApi = 'https://graph.microsoft.com/v1.0/me/photo/$value';

    $params = [
      'headers' => [
        'Authorization' => 'Bearer ' . $token,
      ],
    ];

    $apiRes = \Drupal::service('azure_oauth_sso.token_service')->apiCall($urlApi, $params, 'GET');

    if ($apiRes !== NULL) {
      $apiResBase64 = base64_encode($apiRes);
      $splitArray = str_split($apiResBase64, 1024);
      $apiResBase64Reconstructed = implode('', $splitArray);
      $decodedApiRes = base64_decode($apiResBase64Reconstructed);

      $directory = 'public://profilePic/';

      if (!file_exists($directory)) {
        mkdir($directory, 0777, TRUE);
      }

      $imageInfo = getimagesizefromstring($decodedApiRes);

      if ($imageInfo === FALSE) {
        return NULL;
      }

      $imageType = $imageInfo['mime'];
      $fileExtension = explode('/', $imageType);
      $fileName = $this->displayName . '.' . $fileExtension[1];

      $file_path = $directory . $fileName;

      if ($this->areImagesDifferent($file_path, $decodedApiRes)) {
        file_put_contents($file_path, $decodedApiRes);
        chmod($file_path, 0644);

        $file_entity = $this->saveFileToDatabase($file_path);
        $this->updateUserProfilePicture($file_entity);
      }
    }
  }

  private function saveFileToDatabase($filePath) {
    $file = File::create(
      [
        'uri' => $filePath,
      ]
    );

    $file->save();

    return $file;
  }

  private function updateUserProfilePicture(File $file) {
    try {
      $user = User::load(\Drupal::currentUser()->id());

      if ($user) {
        if ($user->hasField('user_picture') && !$user->get('user_picture')->isEmpty()) {
          $fid = $user->get('user_picture')->target_id;

          $oldFile = File::load($fid);

          if ($oldFile) {
            $file->delete();
          }
        }

        $user->set('user_picture', $file->id());
        $user->save();
      }
    }
    catch (\Exception $e) {
      \Drupal::logger('azure_oauth_sso')->error('An error occurred: @message', ['@message' => $e->getMessage()]);
    }
  }

  private function areImagesDifferent($file_path, $newImageData) {
    try {
      $user = User::load(\Drupal::currentUser()->id());
      $entityId = $user->get('user_picture')->target_id;

      if ($entityId) {
        $file = File::load($entityId);
        if ($file) {
          $fileContents = file_get_contents($file->getFileUri());
          if ($fileContents !== FALSE) {
            return md5($fileContents) !== md5($newImageData);
          }
          else {
            \Drupal::logger('azure_oauth_sso')->error('Failed to read file contents.');
            return TRUE;
          }
        }
        else {
          \Drupal::logger('oauthazure_oauth_sso ')->error('Failed to load file entity.');
          return TRUE;
        }
      }

      return TRUE;
    }
    catch (\Exception $e) {
      \Drupal::logger('azure_oauth_sso')->error('An error occurred: @message', ['@message' => $e->getMessage()]);
      return TRUE;
    }
  }
}
